/**************************************************************************
IAR +STM8S_StdPeriph_Lib_V2.1
Chip type               : STM8S003K3
MCU Core Clock frequency: 16 MHz
Project: VA_meter |7seg led |4dgt
****************************************************************************/
#include "stm8s.h"
#include "iostm003.c"
#include <stdbool.h>
#include "bitset.c"

#define crystal_frq 16E6

#define A       PD_ODR_ODR0
#define B       PD_ODR_ODR2
#define C       PD_ODR_ODR5
#define D       PD_ODR_ODR7
#define E       PD_ODR_ODR6
#define F       PD_ODR_ODR1
#define G       PD_ODR_ODR4
#define DP      PD_ODR_ODR3
#define N0      PC_ODR_ODR4
#define N0_ddr  PC_DDR_DDR4
#define N0_idr  PC_IDR_IDR4
#define N1      PC_ODR_ODR1
#define N2      PC_ODR_ODR2
#define N3      PC_ODR_ODR3

#define ssg_on  TIM1->BKR=TIM1_BKR_MOE
#define ssg_off TIM1->EGR|=TIM1_EGR_BG

enum {CA,CC} ssg_pl;
enum {prs, rls, hold} btn[1];

u8 dbuf[4]={0,1,2,3};
u8 dpp=0;
u8 br_tmp;
enum {start, Volts, Amps} dev_st=start;

u16 adc3_val, adc2_val, adc1_val, adc0_val;
bool adc_ready=0;

u16 tim0,tim1,tim2,tim3,tbr;
u8 tmr0;
bool iwdg_bit;


__no_init static u32 eeR1 @ 0x4000;
__no_init static u32 eeR2 @ 0x4000+sizeof(u32)*1;
__no_init static u32 eeR3 @ 0x4000+sizeof(u32)*2;
__no_init static u32 eeR4 @ 0x4000+sizeof(u32)*3;
__no_init static u32 eeR5 @ 0x4000+sizeof(u32)*4;
__no_init static u32 eeR6 @ 0x4000+sizeof(u32)*5;
__no_init static u32 eeR7 @ 0x4000+sizeof(u32)*6;

__no_init static u8 eeV_br    @ 0x4000+sizeof(u32)*7;
__no_init static u8 ee_new    @ 0x4000+sizeof(u32)*7+sizeof(u8)*1;
__no_init static u8 ee_devst  @ 0x4000+sizeof(u32)*7+sizeof(u8)*2;
__no_init static u8 ee_ref    @ 0x4000+sizeof(u32)*7+sizeof(u8)*3;
__no_init static u8 ee_358offset @ 0x4000+sizeof(u32)*7+sizeof(u8)*4;

void gpio_cfg()
{
  PA_ODR= 1<<1;
  PB_ODR= 0;
  PC_ODR= 0;
  PD_ODR= 0;
  PE_ODR= 0;
  PF_ODR= 0;

  PA_CR1= 1<<1;
  PB_CR1= 1<<7|1<<6;
  PC_CR1= 0xff;
  PD_CR1= 0xff;
  PE_CR1= 0;
  PF_CR1= 0xff;

  PA_DDR= 1<<3|1<<1;
  PB_DDR= 1<<5|1<<4;
  PC_DDR= 0xff;
  PD_DDR= 0xff;
  PE_DDR= 0;
  PF_DDR= 0xff;
}
void clk_cfg()
{
 CLK->CKDIVR = 0;
 CLK->PCKENR1 = CLK_PCKENR1_TIM1|CLK_PCKENR1_TIM2;
 CLK->PCKENR2 = CLK_PCKENR2_ADC;

}
void t1_cfg()
{
 {
  #define t1_rate 1000

   u16 x=(u16)((u32)crystal_frq/0x7fff);

   TIM1->PSCRH = x>>8;
   TIM1->PSCRL = (u8)x;
   x=0x7fff/t1_rate;
   TIM1->ARRH  = x>>8;
   TIM1->ARRL  = (u8)x;

   TIM1->CCR1H  = x>>8;
   TIM1->CCR1L  = (u8)x;
   TIM1->CCR2H  = x>>8;
   TIM1->CCR2L  = (u8)x;
   TIM1->CCR3H  = x>>8;
   TIM1->CCR3L  = (u8)x;
   TIM1->CCR4H  = x>>8;
   TIM1->CCR4L  = (u8)x;

   TIM1->CR2    = TIM1_CR2_CCPC;
  }

  TIM1->CR1 = TIM1_CR1_CEN;
}
void t2_cfg()
{
#define t2_interrupt_rate 500
 u32 x=1;

 for(u8 i=1;i<0x0f+1;i++)
 {
  x=x*2;
  if(crystal_frq/x<0xffff){TIM2->PSCR=i; break;}
 }

 TIM2->ARRH  = (u32)crystal_frq/(x*t2_interrupt_rate)-1>>8;
 TIM2->ARRL  = (u8)(crystal_frq/(x*t2_interrupt_rate)-1);

 TIM2->IER   = 1;
 TIM2->CR1   = 1;
}
void adc_cfg()
{
  ADC1->CR1  = ADC1_PRESSEL_FCPU_D18;
  ADC1->CR2  = ADC1_ALIGN_RIGHT|ADC1_CR2_SCAN;
  ADC1->CR3  = ADC1_CR3_DBUF;
  ADC1->CSR  = 3;
  ADC1->CR1 |= ADC1_CR1_ADON;
  ADC1->TDRL = 0x0f;
  ADC1->CR1 |= ADC1_CR1_ADON;
}
void iwdg_cfg()
{
IWDG_WriteAccessCmd(IWDG_WriteAccess_Enable);
IWDG_SetPrescaler(IWDG_Prescaler_128);
IWDG_SetReload(0);    //2ms
IWDG_Enable();
IWDG_ReloadCounter();
}

void convert_data (u16 d)
{
 if(d>999)
 for (u8 k=0;k<4;k++)
  {
   dbuf[k]=d%10;
   d/=10;
  }
 else
 {
 for (u8 k=0;k<3;k++)
  {
   dbuf[k]=d%10;
   d/=10;
  }
 if(dev_st==Volts)dbuf[3]='u';
 else dbuf[3]='c';
 }
}
void display_data()
{
  static u8 pc;

  (pc)?(pc--):(pc = 3);
  TIM1->CCER1=TIM1->CCER2=0;TIM1->EGR=TIM1_EGR_COMG;

  switch (dbuf[pc])
 {
  case 0: A=0; B=0; C=0; D=0; E=0; F=0; G=1; break;
  case 1: A=1; B=0; C=0; D=1; E=1; F=1; G=1; break;
  case 2: A=0; B=0; C=1; D=0; E=0; F=1; G=0; break;
  case 3: A=0; B=0; C=0; D=0; E=1; F=1; G=0; break;
  case 4: A=1; B=0; C=0; D=1; E=1; F=0; G=0; break;
  case 5: A=0; B=1; C=0; D=0; E=1; F=0; G=0; break;
  case 6: A=0; B=1; C=0; D=0; E=0; F=0; G=0; break;
  case 7: A=0; B=0; C=0; D=1; E=1; F=1; G=1; break;
  case 8: A=0; B=0; C=0; D=0; E=0; F=0; G=0; break;
  case 9: A=0; B=0; C=0; D=0; E=1; F=0; G=0; break;

  case 10: A=0; B=1; C=1; D=1; E=1; F=1; G=1; break;
  case 11: A=1; B=0; C=1; D=1; E=1; F=1; G=1; break;
  case 12: A=1; B=1; C=0; D=1; E=1; F=1; G=1; break;
  case 13: A=1; B=1; C=1; D=0; E=1; F=1; G=1; break;
  case 14: A=1; B=1; C=1; D=1; E=0; F=1; G=1; break;
  case 15: A=1; B=1; C=1; D=1; E=1; F=0; G=1; break;

  case ' ': A=1; B=1; C=1; D=1; E=1; F=1; G=1; break;
  case 'v':
  case 'u': A=1; B=1; C=0; D=0; E=0; F=1; G=1; break;
  case 'c': A=1; B=1; C=1; D=0; E=0; F=1; G=0; break;
  case 'b': A=1; B=1; C=0; D=0; E=0; F=0; G=0; break;
  case 't': A=1; B=1; C=1; D=0; E=0; F=0; G=0; break;
  case 'l': A=1; B=1; C=1; D=0; E=0; F=1; G=1; break;
  case 'o': A=1; B=1; C=0; D=0; E=0; F=1; G=0; break;
  case 'A': A=0; B=0; C=0; D=1; E=0; F=0; G=0; break;
  case 'H': A=1; B=0; C=0; D=1; E=0; F=0; G=0; break;
  case 'C': A=0; B=1; C=1; D=0; E=0; F=0; G=1; break;
  case 'P': A=0; B=0; C=1; D=1; E=0; F=0; G=0; break;
  case 'E': A=0; B=1; C=1; D=0; E=0; F=0; G=0; break;
  case 'U': A=1; B=0; C=0; D=0; E=0; F=0; G=1; break;
  case 'F': A=0; B=1; C=1; D=1; E=0; F=0; G=0; break;
  case 'L': A=1; B=1; C=1; D=0; E=0; F=0; G=1; break;
  case 'r': A=1; B=1; C=1; D=1; E=0; F=1; G=0; break;
  case '_': A=1; B=1; C=1; D=0; E=1; F=1; G=1; break;
  case 'd': A=1; B=0; C=0; D=0; E=0; F=1; G=0; break;
  case 'n': A=1; B=1; C=0; D=1; E=0; F=1; G=0; break;
  case 'G': A=0; B=1; C=0; D=0; E=0; F=0; G=1; break;
  case 'O': A=0; B=0; C=0; D=0; E=0; F=0; G=1; break;
  case 'i': A=1; B=1; C=1; D=1; E=0; F=1; G=1; break;
  case 'I': A=1; B=1; C=1; D=1; E=0; F=0; G=1; break;
  case 'S': A=0; B=1; C=0; D=0; E=1; F=0; G=0; break;
  case 'J': A=1; B=0; C=0; D=0; E=1; F=1; G=1; break;
  case 'h': A=1; B=1; C=0; D=1; E=0; F=0; G=0; break;
  default:  A=1; B=1; C=1; D=1; E=1; F=1; G=0; break; //-
 }
 if(ssg_pl==CC)
 {
  A=~A; B=~B; C=~C; D=~D; E=~E; F=~F; G=~G;
  DP=0;
  switch (pc)
  {
   case 0: TIM1->CCER2 = TIM1_CCER2_CC4E; if(dpp == 0) {DP=1;} break;
   case 1: TIM1->CCER1 = TIM1_CCER1_CC1E; if(dpp == 1) {DP=1;} break;
   case 2: TIM1->CCER1 = TIM1_CCER1_CC2E; if(dpp == 2) {DP=1;} break;
   case 3: if(dbuf[3]){TIM1->CCER2 = TIM1_CCER2_CC3E;} if(dpp == 3) {DP=1;} break;
   default: DP=0; break;
  }
 }
 else
 {
  DP=1;
  switch (pc)
  {
   case 0: TIM1->CCER2 = TIM1_CCER2_CC4E; if(dpp == 0) {DP=0;} break;
   case 1: TIM1->CCER1 = TIM1_CCER1_CC1E; if(dpp == 1) {DP=0;} break;
   case 2:
    if(dev_st==Volts)TIM1->CCER1 = TIM1_CCER1_CC2E;
    else if (dbuf[2])TIM1->CCER1 = TIM1_CCER1_CC2E;
    if(dpp == 2) {DP=0;}
    break;
   case 3: if(dbuf[3]){TIM1->CCER2 = TIM1_CCER2_CC3E;} if(dpp == 3) {DP=0;} break;
   default: DP=1; break;
  }
 }
 TIM1->EGR = TIM1_EGR_COMG;
}

void scan_keys ()
{
#define key0 PB_IDR_IDR7

#define lp  25
#define lh  30
#define lr  5
#define noc 1

static u8 count[noc*3];
u8 key_st[noc];

key_st[0] = key0;

 for(u8 i=0;i<noc;i++)
 {
  if (!key_st[i]) {count[i*3]++;count[i*3+2]=lr;}
  else if((key_st[i])&&(!count[i*3+2])) {count[i*3]=count[i*3+1]=0;btn[i]=rls;}
  else count[i*3+2]--;

  if ((count[i*3]>lp)&&(btn[i]!=hold)) {btn[i]=prs; count[i*3]=0; count[i*3+1]++;}
  if (count[i*3+1]>lh) {btn[i]=hold; count[i*3+1]=0;}
 }
}
void scan_adc()
{
 static u32 adc3_cps=0, adc2_cps=0, adc1_cps=0, adc0_cps=0;
 static u8 adc_cnt=0;

  adc3_cps+= *(u16*)&ADC1->DB3RH;
  adc2_cps+= *(u16*)&ADC1->DB2RH;
  adc1_cps+= *(u16*)&ADC1->DB1RH;
  adc0_cps+= *(u16*)&ADC1->DB0RH;
  ADC1->CSR = 3;
  ADC1->CR1|=ADC1_CR1_ADON;

 #define vac 250
  if(adc_cnt<vac){adc_cnt++;}
  else
  {
   adc3_val = (u16)(adc3_cps/(vac+1));
   adc2_val = (u16)(adc2_cps/(vac+1));
   adc1_val = (u16)(adc1_cps/(vac+1));
   adc0_val = (u16)(adc0_cps/(vac+1));
   adc_cnt=adc3_cps=adc2_cps=adc1_cps=adc0_cps = 0;
   adc_ready=1;
  }
}
void detect_7seg_polarity()
{
 u8 n=0;

 N3=N2=N1=N0=0;
 ssg_off;
 A=B=C=D=E=F=G=DP=0;
 N0_ddr=0;
 for(volatile u8 i=0xff;i;i--)
 {
  if(!N0_idr)n++;
  else if(n)n--;
 }
  if(n>0xff/2)
  {
   ssg_pl=CA;
   TIM1->CCMR1 = TIM1_OCMODE_PWM1;
   TIM1->CCMR2 = TIM1_OCMODE_PWM1;
   TIM1->CCMR3 = TIM1_OCMODE_PWM1;
   TIM1->CCMR4 = TIM1_OCMODE_PWM1;
  }
  else
  {
   ssg_pl=CC;
   N3=N2=N1=N0=1;
   TIM1->CCMR1 = TIM1_OCMODE_PWM2;
   TIM1->CCMR2 = TIM1_OCMODE_PWM2;
   TIM1->CCMR3 = TIM1_OCMODE_PWM2;
   TIM1->CCMR4 = TIM1_OCMODE_PWM2;
  }

 N0_ddr=1;
}
void set_brightness(u8 bt)
{
 u16 t1t;

 if(bt>100)bt=100;

   t1t = (u16)TIM1->ARRH<<8;
   t1t+= TIM1->ARRL;
   t1t = t1t*bt/100;

   TIM1->CCR1H  = t1t>>8;
   TIM1->CCR1L  = (u8)t1t;
   TIM1->CCR2H  = TIM1->CCR1H;
   TIM1->CCR2L  = TIM1->CCR1L;
   TIM1->CCR3H  = TIM1->CCR1H;
   TIM1->CCR3L  = TIM1->CCR1L;
   TIM1->CCR4H  = TIM1->CCR1H;
   TIM1->CCR4L  = TIM1->CCR1L;
}

void devcon();

INTERRUPT_HANDLER(TIM2_OVF, 13)
{
 scan_adc();
 scan_keys();
 display_data();

 if(tim0)tim0--;
 if(tim1)tim1--;
 if(tim2)tim2--;
 if(tim3)tim3--;
 if(tmr0)tmr0--;
 if(tbr)tbr--;

 TIM2->SR1 = 0;
 if(iwdg_bit){iwdg_bit=0; IWDG->KR = IWDG_KEY_REFRESH;}
}


void main()
{
 gpio_cfg();
 clk_cfg();
 adc_cfg();
 t1_cfg();
 t2_cfg();

 if(ee_new!=0xfa)
 {
 FLASH_Unlock(FLASH_MEMTYPE_DATA);
  eeR1   = 30000; //Ohm
  eeR2   = 10000; //Ohm
  eeR3   = 100000;//Ohm
  eeR4   = 10000; //Ohm
  eeR5   = 10000; //uOhm
  eeR6   = 30000; //Ohm
  eeR7   = 1000;  //Ohm
  eeV_br   = 100;
  ee_devst = 0;
  ee_ref   = 181; //Vref*100
  ee_358offset = 5; //mV

  ee_new   = 0xfa;
 FLASH_Lock(FLASH_MEMTYPE_DATA);
 }

 detect_7seg_polarity();
 br_tmp=eeV_br;
 ssg_on;

 iwdg_cfg();

 asm("rim");

  while(1)
  {
   devcon();
   iwdg_bit=1;

  }

}


void devcon()
{
 #define ref_adc adc3_val //~561

 static bool itr=0;
 static bool a_trg, b_trg, c_trg, d_trg=0;
 static u8 br_tmp1=0;
 static u32 r1,r2,r3,r4,r5,r6,r7;
 static u8 ref, offset;


 if(btn[0]==rls){a_trg=0;if(b_trg){b_trg=0;c_trg^=1;}}

// brightness adj
 {
  if(btn[0]==hold)
  {
   b_trg=1;

   if(!tmr0&&c_trg)
   {
    tmr0=20;
    if(br_tmp<100)br_tmp++;
   }
   else if(!tmr0)
   {
    tmr0=20;
    if(br_tmp>10)br_tmp--;
   }
  }

  if(br_tmp1!=br_tmp){br_tmp1=br_tmp;set_brightness(br_tmp);tbr=5000;}

  if(!tbr&&br_tmp!=eeV_br)
  {
   FLASH_Unlock(FLASH_MEMTYPE_DATA);
   eeV_br=br_tmp;
   FLASH_Lock(FLASH_MEMTYPE_DATA);
  }
 }
 //end of brightness adj

 switch(dev_st)
 {
  case start:
  {
   if(!itr)
   {
	 itr=1;
    if(!ee_devst) dev_st=Volts; else dev_st=Amps;

    r1 = eeR1;
    r2 = eeR2;
    r3 = eeR3;
    r4 = eeR4;
    r5 = eeR5;
    r6 = eeR6;
    r7 = eeR7;
    ref = ee_ref;
    offset = ee_358offset;
   }

   for(u8 i=0;i<7;)
   {
    iwdg_bit=1;
    if(!tim0)
    {
     tim0=50;
     if(i<6){dbuf[3]=dbuf[2]=dbuf[1]=dbuf[0]=10+i; dpp=9;}
     else dbuf[3]=dbuf[2]=dbuf[1]=dbuf[0]=' ';
     i++;
    }
   }

  if(dev_st!=start){itr=0;}
  }
  break;

  case Volts:
  {
   #define gain (r6*1000/r7+1*1000)
   #define shunt r5
   #define Imax_adc (u16)((u32)ref*0x3ff*10000/ref_adc*100/gain*1000/shunt) //~10(.)64A
   #define VL_max (u16)((u32)ref*0x3ff/10*(r1+r2)/ref_adc*10/r2) //~13(.)20V max
   #define VH_max (u16)((u32)ref*0x3ff/10*(r3+r4)/ref_adc*10/r4) //~36(.)30V max

   if(!itr)
   {
	 itr=1;
    dbuf[3]='u';
    dpp=2;
    tim3=50000;
   }
//mode switching
   if(btn[0]==hold){d_trg=0; a_trg=0;}
   else if(btn[0]==prs&&!d_trg&&!a_trg){d_trg=1; a_trg=1;}
   else if(!a_trg&&d_trg){dev_st=Amps; d_trg=0;}
//end of mode switching


   if(adc_ready)
   {
    adc_ready=0;
    if(!ref_adc)ref_adc=1;
    if (adc0_val < 1020)
    {
     if((u16)((u32)adc0_val*VL_max/0x3ff)>(u16)((u32)adc2_val*Imax_adc*shunt/(0x3ffUL*1000000)))
     convert_data((u16)((u32)adc0_val*VL_max/0x3ff)-(u16)((u32)adc2_val*Imax_adc*shunt/(0x3ffUL*1000000)));
     else convert_data(0);
    }
    else
     convert_data((u16)((u32)adc1_val*VH_max/0x3ff)-(u16)((u32)adc2_val*Imax_adc*shunt/(0x3ffUL*1000000)));
   }

   if(!tim3&&ee_devst)
   {
   FLASH_Unlock(FLASH_MEMTYPE_DATA);
    ee_devst=0;
   FLASH_Lock(FLASH_MEMTYPE_DATA);
   }
  if(dev_st!=Volts){itr=0;}
  }
  break;

  case Amps:
  {
   if(!itr)
   {
	 itr=1;
    dbuf[3]='c';dpp=1;
    tim3=50000;
   }

   if(adc_ready)
   {
    adc_ready=0;
    if((u16)(adc2_val>(u32)offset*gain/ref*1023*ref_adc/10000000UL))
    convert_data((u16)((u32)adc2_val*Imax_adc/(0x3ff*10)));
    else convert_data(0);
   }

//mode switching
   if(btn[0]==hold){d_trg=0; a_trg=0;}
   else if(btn[0]==prs&&!d_trg&&!a_trg){d_trg=1; a_trg=1;}
   else if(!a_trg&&d_trg){dev_st=Volts; d_trg=0;}
//end of mode switching

   if(!tim3&&!ee_devst)
   {
   FLASH_Unlock(FLASH_MEMTYPE_DATA);
    ee_devst=1;
   FLASH_Lock(FLASH_MEMTYPE_DATA);
   }
  if(dev_st!=Amps){itr=0;}
  }
  break;

 default:break;
 }

}

